/* KallistiGL for KOS 1.1.7

   gldraw.c
   (c)2001 Dan Potter
*/

#include <GL/gl.h>
#include <GL/glu.h>
#include "glinternal.h"
#include "glmacros.h"

#include <assert.h>
#include <string.h>
#include <dcutils/3dutils.h>
#include <dc/pvr.h>

CVSID("$Id: gldraw.c,v 1.19 2002/04/07 04:57:30 bardtx Exp $");

/* internal prototypes */
static void send_user_clip();

/* Drawing functions */

#define VBUF_SIZE 128 

/* It is important to align the vertex buffers for use with the
   the direct render API. I found 8192 gave the best results, but 32,
   64, and 1024 also worked well. To see cache thrashing in action try
   setting the alignment to 16384 or 32768. The polys per second will
   drop by about half. 
   
   mat_transform seems to work best when the source and destination buffers
   are seperate.  Therefore we have a pre and post transform vertex buffer */

/* Pre-xformed vertex buffer */
static pvr_vertex_t vtxbuf[VBUF_SIZE] __attribute__((aligned(8192)));
/* Post-xformed vertex buffer */
static pvr_vertex_t xfrmbuf[VBUF_SIZE] __attribute__((aligned(8192)));
static int vbuf_top;

/* This portion of the pipeline will change drastically */

/* Current vertex attributes (move to global?) */
static GLuint vert_rgba = 0xffffffff;
static GLfloat vert_u = 0.0f, vert_v = 0.0f;

/* Make sure the proper matrices are in the matrix registers */
static void fix_matrices() {
	if (gl_matrix_dirty) {
		mat_identity();
		glKosMatrixApply(GL_KOS_SCREENVIEW);
		glKosMatrixApply(GL_PROJECTION);
		glKosMatrixApply(GL_MODELVIEW);

		gl_matrix_dirty = GL_FALSE;
	}
}

/* Negative 1/w values annoy the pvr, so check for them */
static inline int check_w(pvr_vertex_t *vert, int n) {
	int i;

	for(i=0; i<n; i++)
	        if (vert[i].z < 0) return 0; 
	return 1;
}

static short xor[2][2] ={
	{1,0}, /* FFCW, FFCCW */ 
	{0,1}  /* BFCW, BFCCW */
};
/* GL_POINTS always visible! */
static void end_points() {
	int i;
	volatile pvr_vertex_t	*vert;
	pvr_dr_state_t	dr_state;

	float size1 = (xor[(gl_cull_mode == GL_BACK)][(gl_front_face == GL_CCW)]) ? gl_point_size : -gl_point_size;
	float size2 = ABS(size1);
	
	pvr_dr_init(dr_state);

	/* Should be one point in the buffer */
	assert_msg(vbuf_top >= 1, "Need at least one point.");

	/* create a point using a quad */
	memcpy(&vtxbuf[1], &vtxbuf[0], sizeof(pvr_vertex_t));
	memcpy(&vtxbuf[2], &vtxbuf[0], sizeof(pvr_vertex_t));
	memcpy(&vtxbuf[3], &vtxbuf[0], sizeof(pvr_vertex_t));

	vtxbuf[3].flags = PVR_CMD_VERTEX_EOL;
	/* mat_transform((vector_t*)&vtxbuf[0].x, (vector_t*)&xfrmbuf[0].x, 4, sizeof(pvr_vertex_t)); */

	for (i=0; i<vbuf_top; i++) {
		float x = vtxbuf[i].x, y = vtxbuf[i].y, z = vtxbuf[i].z, w = 1.0f;
		mat_trans_single4(x, y, z, w);
		xfrmbuf[i].x = x;
		xfrmbuf[i].y = y;
		if (w == 1.0f)
			xfrmbuf[i].z = ((gl_viewport_scale[2] * z) + gl_viewport_offset[2]);
		else
			xfrmbuf[i].z = w;
	}

	xfrmbuf[0].x += size1;
	xfrmbuf[0].y += size2;
	xfrmbuf[1].x += size1;
	xfrmbuf[1].y -= size2;
	xfrmbuf[2].x -= size1;
	xfrmbuf[2].y += size2;
	xfrmbuf[3].x -= size1;
	xfrmbuf[3].y -= size2;

	/* Throw it all to the TA using direct render */
	if (check_w(xfrmbuf, 4))
		for (i=0; i<4; i++) {
			vert = pvr_dr_target(dr_state);
			vert->flags = vtxbuf[i].flags;
			vert->x     = xfrmbuf[i].x;
			vert->y     = xfrmbuf[i].y;
			vert->z     = xfrmbuf[i].z;
			vert->u     = vtxbuf[i].u;
			vert->v     = vtxbuf[i].v;
			vert->argb  = vtxbuf[i].argb;
			vert->oargb = vtxbuf[i].oargb;
			pvr_dr_commit(vert);
		}
}

static void end_triangles() {
	int i;
	volatile pvr_vertex_t	*vert;
	pvr_dr_state_t	dr_state;
	
	pvr_dr_init(dr_state);

	/* Should be three points in the buffer */
	assert_msg(vbuf_top == 3, "Wrong number of vertices.");

	vtxbuf[2].flags = PVR_CMD_VERTEX_EOL;
	/* mat_transform((vector_t*)&vtxbuf[0].x, (vector_t*)&xfrmbuf[0].x, 3, sizeof(pvr_vertex_t)); */

	for (i=0; i<vbuf_top; i++) {
		float x = vtxbuf[i].x, y = vtxbuf[i].y, z = vtxbuf[i].z, w = 1.0f;
		mat_trans_single4(x, y, z, w);
		xfrmbuf[i].x = x;
		xfrmbuf[i].y = y;
		if (w == 1.0f)
			xfrmbuf[i].z = ((gl_viewport_scale[2] * z) + gl_viewport_offset[2]);
		else
			xfrmbuf[i].z = w;
	}

	/* Throw it all to the TA using direct render */
	if (check_w(xfrmbuf, 3))
		for (i=0; i<3; i++) {
			vert = pvr_dr_target(dr_state);
			vert->flags = vtxbuf[i].flags;
			vert->x     = xfrmbuf[i].x;
			vert->y     = xfrmbuf[i].y;
			vert->z     = xfrmbuf[i].z;
			vert->u     = vtxbuf[i].u;
			vert->v     = vtxbuf[i].v;
			vert->argb  = vtxbuf[i].argb;
			vert->oargb = vtxbuf[i].oargb;
			pvr_dr_commit(vert);
		}
}

static void end_triangle_strip() {
	int i;
	volatile pvr_vertex_t	*vert;
	pvr_dr_state_t	dr_state;
	
	pvr_dr_init(dr_state);

	/* Should be at least 3 points in the buffer */
	assert_msg(vbuf_top >= 3, "Need at least three vertices.");

	vtxbuf[vbuf_top-1].flags = PVR_CMD_VERTEX_EOL;
	/* mat_transform((vector_t*)&vtxbuf[0].x, (vector_t*)&xfrmbuf[0].x, vbuf_top, sizeof(pvr_vertex_t)); */

	for (i=0; i<vbuf_top; i++) {
		float x = vtxbuf[i].x, y = vtxbuf[i].y, z = vtxbuf[i].z, w = 1.0f;
		mat_trans_single4(x, y, z, w);
		xfrmbuf[i].x = x;
		xfrmbuf[i].y = y;
		if (w == 1.0f)
			xfrmbuf[i].z = ((gl_viewport_scale[2] * z) + gl_viewport_offset[2]);
		else
			xfrmbuf[i].z = w;
	}

	/* Throw it all to the TA using direct render */
	if (check_w(xfrmbuf, vbuf_top))
		for (i=0; i<vbuf_top; i++) {
			vert = pvr_dr_target(dr_state);
			vert->flags = vtxbuf[i].flags;
			vert->x     = xfrmbuf[i].x;
			vert->y     = xfrmbuf[i].y;
			vert->z     = xfrmbuf[i].z;
			vert->u     = vtxbuf[i].u;
			vert->v     = vtxbuf[i].v;
			vert->argb  = vtxbuf[i].argb;
			vert->oargb = vtxbuf[i].oargb;
			pvr_dr_commit(vert);
		}
}

static void end_quads() {
	int i, j, order[4] = {3, 0, 2, 1};
	volatile pvr_vertex_t	*vert;
	pvr_dr_state_t	dr_state;
	
	pvr_dr_init(dr_state);

	/* Should be four points in the buffer */
	assert_msg(vbuf_top == 4, "Wrong number of vertices.");

	vtxbuf[order[3]].flags = PVR_CMD_VERTEX_EOL;
	/* mat_transform((vector_t*)&vtxbuf[0].x, (vector_t*)&xfrmbuf[0].x, 4, sizeof(pvr_vertex_t)); */

	for (i=0; i<vbuf_top; i++) {
		float x = vtxbuf[i].x, y = vtxbuf[i].y, z = vtxbuf[i].z, w = 1.0f;
		mat_trans_single4(x, y, z, w);
		xfrmbuf[i].x = x;
		xfrmbuf[i].y = y;
		if (w == 1.0f)
			xfrmbuf[i].z = ((gl_viewport_scale[2] * z) + gl_viewport_offset[2]);
		else
			xfrmbuf[i].z = w;
	}

	/* Throw it all to the TA using direct render */
	if (check_w(xfrmbuf, 4))
		for (i=0; i<4; i++) {
			j = order[i];
			vert = pvr_dr_target(dr_state);
			vert->flags = vtxbuf[j].flags;
			vert->x     = xfrmbuf[j].x;
			vert->y     = xfrmbuf[j].y;
			vert->z     = xfrmbuf[j].z;
			vert->u     = vtxbuf[j].u;
			vert->v     = vtxbuf[j].v;
			vert->argb  = vtxbuf[j].argb;
			vert->oargb = vtxbuf[j].oargb;
			pvr_dr_commit(vert);
		}
}

static void check_end() {
	/* What type of primitive was it? */
	switch(gl_prim_type) {
	case GL_POINTS:
		end_points();
		vbuf_top = 0;
		break;
	case GL_TRIANGLES:
		if (vbuf_top == 3) {
			end_triangles();
			vbuf_top = 0;
		}
		break;
	case GL_TRIANGLE_STRIP:
		break;
	case GL_QUADS:
		if (vbuf_top == 4) {
			end_quads();
			vbuf_top = 0;
		}
		break;
	default:
		assert_msg(0, "gl_prim_type: Unknown primitive type.");
	}
}

static pvr_poly_hdr_t polyhdr;

void send_poly_hdr(void) {

	if (TXR_ENABLED) {
		memcpy(&gl_poly_cxt.txr, &gl_cur_texture->txr, sizeof(gl_poly_cxt.txr));
	}

	switch(gl_active_list) {
	case GL_LIST_OPAQUE_POLY:
		gl_poly_cxt.gen.alpha = PVR_ALPHA_DISABLE;
		gl_poly_cxt.txr.alpha = PVR_TXRALPHA_ENABLE;
		gl_poly_cxt.blend.src = PVR_BLEND_ONE;
		gl_poly_cxt.blend.dst = PVR_BLEND_ZERO;
		break;
	case GL_LIST_OPAQUE_MOD:
		break;
	case GL_LIST_TRANS_POLY:
	case GL_LIST_PUNCHTHRU:
		gl_poly_cxt.gen.alpha = PVR_ALPHA_ENABLE;
		gl_poly_cxt.txr.alpha = PVR_TXRALPHA_ENABLE;
 		gl_poly_cxt.blend.src = gl_blend_src;
		gl_poly_cxt.blend.dst = gl_blend_dst;
		break;
	case GL_LIST_TRANS_MOD:
		break;
	}
	/* Set states that we couldn't before */
	if (gl_cull_face)
	{
		if (gl_cull_mode == GL_BACK)
			if (gl_front_face == GL_CW)
				gl_poly_cxt.gen.culling = PVR_CULLING_CCW;
			else
				gl_poly_cxt.gen.culling = PVR_CULLING_CW;
		else
			if (gl_front_face == GL_CCW)
				gl_poly_cxt.gen.culling = PVR_CULLING_CCW;
			else
				gl_poly_cxt.gen.culling = PVR_CULLING_CW;
	} else
		gl_poly_cxt.gen.culling = PVR_CULLING_NONE;

	

	pvr_poly_compile(&polyhdr, &gl_poly_cxt);
	
	/* if this bit isn't set when texturing is disabled, blue
	   doesn't work for CW polygons? [BM]*/
	if (!TXR_ENABLED)
		polyhdr.mode3 |= 0x20000000;

//	printf("poly header: %08x %08x %08x %08x\n", polyhdr.cmd, polyhdr.mode1, polyhdr.mode2, polyhdr.mode3);

	pvr_prim(&polyhdr, sizeof(pvr_poly_hdr_t));
}

/* This can be used for KGL Direct Rendering */
void glKosPolyHdrSend() {
	send_poly_hdr();
}

void glBegin(GLenum mode) {
	/* Submit the user clip rectangle (if necessary) */
	if (gl_scissor_dirty) {
		send_user_clip();
		gl_scissor_dirty = GL_FALSE;
	}
	
	/* Submit the poly buffer (if necessary) */
	if (!gl_pbuf_submitted) {
		send_poly_hdr();
		gl_pbuf_submitted = GL_TRUE;
	}
	fix_matrices();

	gl_prim_type = mode;
	vbuf_top = 0;
}

void glEnd(void) {
	/* What type of primitive was it? */
	switch(gl_prim_type) {
	case GL_POINTS:
		break;
	case GL_TRIANGLES:
		break;
	case GL_TRIANGLE_STRIP:
		end_triangle_strip();
		break;
	case GL_QUADS:
		break;
	default:
		assert_msg(0, "gl_prim_type: Unknown primitive type.");
	}
	gl_prim_type = 0;
}

static float autouv[4][2] = {
	{0.0f, 0.0f},
	{1.0f, 0.0f},
	{1.0f, 1.0f},
	{0.0f, 1.0f}
};

void glVertex3f(GLfloat x, GLfloat y, GLfloat z) {
	vtxbuf[vbuf_top].flags = PVR_CMD_VERTEX;

	vtxbuf[vbuf_top].x = x;
	vtxbuf[vbuf_top].y = y;
	vtxbuf[vbuf_top].z = z;
	if (AUTOUV_ENABLED) {
		vert_u = autouv[vbuf_top][0];
		vert_v = autouv[vbuf_top][1];
	}
	/* XXX - Must set uv values to zero if texture is disabled!? */
	if (TXR_ENABLED) {
		vtxbuf[vbuf_top].u = vert_u;
		vtxbuf[vbuf_top].v = vert_v;
	} else {
		vtxbuf[vbuf_top].u = 0.0f;
		vtxbuf[vbuf_top].v = 0.0f;
	}
	vtxbuf[vbuf_top].argb = vert_rgba;
	vtxbuf[vbuf_top].oargb = 0xff000000;

	vbuf_top++;
	assert_msg(vbuf_top < VBUF_SIZE, "Vertex buffer overflow");
	check_end();
}

void glVertex3fv(GLfloat *v) {
        glVertex3f(*v, *(v+1), *(v+2));
}

void glNormal3f(GLfloat nx, GLfloat ny, GLfloat nz) {
}

void glColor4ub(GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha) {
	vert_rgba = (alpha << 24) | (red << 16) | (green << 8) | blue;
}

void glColor3f(GLfloat red, GLfloat green, GLfloat blue) {
	vert_rgba = (0xff << 24) |
		((uint32)(red * 0xff) << 16) |
		((uint32)(green * 0xff) << 8) |
		((uint32)(blue * 0xff));
}

void glColor3fv(GLfloat *v) {
        glColor3f(*v, *(v+1), *(v+2));
}

void glColor4f(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha) {
	vert_rgba = ((uint32)(alpha * 0xff) << 24) |
		((uint32)(red * 0xff) << 16) |
		((uint32)(green * 0xff) << 8) |
		((uint32)(blue * 0xff));
}

void glColor4fv(GLfloat *v) {
        glColor4f(*v, *(v+1), *(v+2), *(v+3));
}

void glTexCoord2f(GLfloat s, GLfloat t) {
	vert_u = s;
	vert_v = t;
}

void glTexCoord2fv(GLfloat *v) {
	vert_u = *v;
	vert_v = *(v+1);
}

static pvr_poly_hdr_t user_clip = {
        PVR_CMD_USERCLIP, 0x00000000, 0x00000000, 0x00000000,
        0x00000000, 0x00000000, 0x00000000, 0x00000000
};

static void send_user_clip() {
	pvr_prim(&user_clip, sizeof(pvr_poly_hdr_t));
}
/* Setup the hardware user clip rectangle. It will be sent during glBegin.

   The minimum clip rectangle is a 32x32 area which is dependent on the tile
   size use by the tile accelerator. The PVR swithes off rendering to tiles
   outside or inside the defined rectangle dependant upon the 'clipmode' bits
   in the polygon header.
   
   Clip rectangles therefore must have a size that is some multiple of 32.
   
   	glScissor(0, 0, 32, 32) allows only the 'tile' in the lower left
	hand corner of the screen to be modified and glScissor(0, 0, 0, 0)
	disallows modification to all 'tiles' on the screen.
*/
void glScissor(GLint x, GLint y, GLsizei width, GLsizei height) {
	GLint minx, miny, maxx, maxy;
	
	assert_msg(!gl_prim_type, "Not allowed within glBegin/glEnd pair.");
	assert((width >= 0) && (height >= 0));
	gl_scissor_x = x;
	gl_scissor_y = y;
	gl_scissor_width = CLAMP(width, 0, gl_screen_width);
	gl_scissor_height = CLAMP(height, 0, gl_screen_height);
	
	minx = gl_scissor_x;
	/* force the origin to the lower left-hand corner of the screen */
	miny = /**/(gl_screen_height - gl_scissor_height) - /**/ gl_scissor_y;
	maxx = (gl_scissor_width + minx);
	maxy = (gl_scissor_height + miny);

	/* load command structure while mapping screen coords to TA tiles */
	user_clip.d1 = CLAMP(minx/32, 0, gl_screen_width/32);
	user_clip.d2 = CLAMP(miny/32, 0, gl_screen_height/32);
	user_clip.d3 = CLAMP((maxx/32)-1, 0, gl_screen_width/32);
	user_clip.d4 = CLAMP((maxy/32)-1, 0, gl_screen_height/32);
	
	gl_scissor_dirty = GL_TRUE;
	gl_pbuf_submitted = GL_FALSE;
}
